from .mapping import get_coord_main_cities, get_plz_mapping
from .visualisation import visualise_plz

__all__ = {
    "plz-bundeland-landkreis mapping": "get_plz_mapping",
    "top_german_cities-coordinates mapping": "get_coord_main_cities",
    "visualise data on germany map with postleitzahl": "visualise_plz",
}



import os

import pandas as pd
import pytest
from sklearn.datasets import make_classification
from sklearn.exceptions import NotFittedError

os.environ["SAM_ML_LOG_LEVEL"] = "debug"
from sam_ml.models import (
    ABC,
    BC,
    BNB,
    DTC,
    ETC,
    GBM,
    GNB,
    GPC,
    KNC,
    LDA,
    LR,
    LSVC,
    MLPC,
    QDA,
    RFC,
    SVC,
    XGBC,
)
from sam_ml.models.main_classifier import SMAC_INSTALLED


def get_models() -> list:
    return [ABC(), BC(), BNB(), DTC(), ETC(), GNB(), GPC(), GBM(), KNC(), LDA(), LSVC(), LR(), MLPC(), QDA(), RFC(), SVC(), XGBC()]

X, Y = make_classification(n_samples = 50,
                            n_features = 5,
                            n_informative = 5,
                            n_redundant = 0,
                            n_classes = 3,
                            weights = [.2, .3, .8])
X = pd.DataFrame(X, columns=["col1", "col2", "col3", "col4", "col5"])
Y = pd.Series(Y)


def test_classifier_fit_evaluate():
    for classifier in get_models():
        classifier.fit(X, Y)
        classifier.evaluate(X, Y, console_out=False)

def test_evaluate_score_error():
    with pytest.raises(NotFittedError):
        for classifier in get_models():
            classifier.evaluate_score(X,Y)

def test_evaluate_score_score():
    for classifier in get_models():
        classifier.train(X, Y)
        classifier.evaluate_score(X,Y)

def test_classifier_train_evaluate():
    for classifier in get_models():
        classifier.train(X, Y, console_out=False)
        classifier.evaluate(X, Y, console_out=False)
        classifier.evaluate_score(X, Y)

def test_classifier_crossvalidation():
    for classifier in get_models():
        classifier.cross_validation(X, Y, cv_num=2)

def test_classifier_crossvalidation_small_data():
    for classifier in get_models():
        classifier.cross_validation_small_data(X, Y)

def test_classifier_randomCVsearch():
    for classifier in get_models():
        best_param, _ = classifier.randomCVsearch(X, Y, n_trails=5, cv_num=2)
        assert best_param != {}, "should always find a parameter combination"

def test_classifier_randomCVsearch_small_data():
    for classifier in get_models():
        best_param, _ = classifier.randomCVsearch(X, Y, n_trails=5, cv_num=2, small_data_eval=True)
        assert best_param != {}, "should always find a parameter combination"

@pytest.mark.with_swig
def test_classifier_smac_search():
    if SMAC_INSTALLED:
        for classifier in get_models():
            best_param = classifier.smac_search(X, Y, n_trails=5, cv_num=2)
            assert best_param != {}, "should always find a parameter combination"
    else:
        with pytest.raises(ImportError):
            for classifier in get_models():
                best_param = classifier.smac_search(X, Y, n_trails=5, cv_num=2)
                assert best_param != {}, "should always find a parameter combination"

@pytest.mark.with_swig
def test_classifier_smac_search_small_data():
    if SMAC_INSTALLED:
        for classifier in get_models():
            best_param = classifier.smac_search(X, Y, n_trails=5, cv_num=2, small_data_eval=True)
            assert best_param != {}, "should always find a parameter combination"
    else:
        with pytest.raises(ImportError):
            for classifier in get_models():
                best_param = classifier.smac_search(X, Y, n_trails=5, cv_num=2, small_data_eval=True)
                assert best_param != {}, "should always find a parameter combination"



import geopandas as gpd
import matplotlib.pyplot as plt
import pandas as pd
from pkg_resources import resource_filename

from .mapping import get_coord_main_cities

plt.style.use('seaborn')

def visualise_plz(plz_region_df: pd.DataFrame, plot_col_name: str, plot_path: str = "german_map.png", plot_title: str = 'Germany map'):
    """
    @param:
        plz_region_df: dataframe with 'plz' column dtype string
        plot_col_name: column to plot
        plot_path: path for saving plot

    @soruce: https://www.suche-postleitzahl.org/downloads, 18/07/2023
    """
    filepath = resource_filename(__name__, 'plz-5stellig.shp')
    plz_shape_df = gpd.read_file(filepath, dtype={'plz': str})
    top_cities = get_coord_main_cities()

    germany_df = pd.merge(
        left=plz_shape_df, 
        right=plz_region_df, 
        on='plz',
        how='inner'
    )
    germany_df.drop(['note'], axis=1, inplace=True)

    plt.rcParams['figure.figsize'] = [16, 11]

    fig, ax = plt.subplots()

    germany_df.plot(
        ax=ax, 
        column=plot_col_name, 
        categorical=True, 
        legend=True, 
        legend_kwds={'title': plot_col_name, 'bbox_to_anchor': (1.35, 0.8)},
        cmap='tab20',
        alpha=0.9,
    )

    for c in top_cities.keys():

        ax.text(
            x=top_cities[c][0], 
            y=top_cities[c][1] + 0.08, 
            s=c, 
            fontsize=12,
            ha='center', 
        )

        ax.plot(
            top_cities[c][0], 
            top_cities[c][1], 
            marker='o',
            c='black', 
            alpha=0.5,
        )

    ax.set(
        title=plot_title,
        aspect=1.3,
        facecolor='white',
    );

    # if default path -> add column name of plot column
    if plot_path == "german_map.png":
        plot_path = plot_path.split(".")[0]+f"_{plot_col_name}."+plot_path.split(".")[1]

    fig.savefig(plot_path)



import pandas as pd

from sam_ml.config import setup_logger

from .sampling import Sampler

logger = setup_logger(__name__)

class SamplerPipeline:
    def __init__(self, algorithm: str | list[Sampler] = "SMOTE_rus_20_50"):
        """
        Class uses multplie up- and down-sampling algorithms instead of only one

        @param:
            - algorithm = "A1_A2_..._An_x1_x2_..._xn": first, use Sampler A1 with sampling_strategy x1% on data, then Sampler A2 with sampling_strategy x2% until Sampler An with sampling_strategy xn on data (only works for binary data!!!)
            - algorithm = list[Sampler]: use each Sampler in list one after the other on data

        @Note:
            - sampling_strategy is the percentage of minority class size of majority class size

        @example:
            - ros_rus_10_50: RandomOverSampler for minority class to 10% of majority class and then RandomUnderSampler for majority class to 2*minority class
            - SMOTE_rus_20_50: SMOTE for minority class to 20% of majority class and then RandomUnderSampler for majority class to 2*minority class
        """
        if type(algorithm) == str:
            self.algorithm = algorithm

            samplers_ratios = algorithm.split("_")
            if len(samplers_ratios)%2 == 1:
                raise ValueError(f"The string has to contain for every Sampler a sampling_strategy, but {samplers_ratios}")
            
            samplers = samplers_ratios[:int(len(samplers_ratios)/2)]
            ratios = samplers_ratios[int(len(samplers_ratios)/2):]
            ratios_float = [int(ratio)/100 for ratio in ratios]

            self.sampler = [Sampler(algorithm=samplers[idx], sampling_strategy=ratios_float[idx]) for idx in range(len(samplers))]
        else:
            self.sampler = algorithm
            self.algorithm = "custom"
    
    def __repr__(self) -> str:
        return f"SamplerPipeline{tuple(self.sampler)}"
    
    @staticmethod
    def check_is_valid_algorithm(algorithm: str) -> bool:
        """
        @return:
            True if algorithm is valid
        """
        samplers_ratios = algorithm.split("_")
        if len(samplers_ratios)%2 == 1:
            logger.warning(f"The string has to contain for every Sampler a sampling_strategy, but {samplers_ratios}")
            return False
        
        samplers = samplers_ratios[:int(len(samplers_ratios)/2)]
        ratios = samplers_ratios[int(len(samplers_ratios)/2):]
        ratios_float = [int(ratio)/100 for ratio in ratios]

        for idx in range(len(samplers)):
            if not (samplers[idx] in Sampler.params()["algorithm"] and 0<ratios_float[idx]<=1):
                logger.warning(f"invalid sampler-sampling_strategy pair: '{samplers[idx]}' with {ratios_float[idx]}")
                return False

        return True
    
    def get_params(self, deep: bool = True):
        return {"algorithm": self.sampler}

    def set_params(self, *params):
        self.sampler = params
        return self

    def sample(self, x_train: pd.DataFrame, y_train: pd.Series) -> tuple[pd.DataFrame, pd.Series]:
        """
        Function for up- and downsampling

        @return:
            tuple x_train_sampled, y_train_sampled
        """
        for sampler_idx in range(len(self.sampler)):
            if sampler_idx == 0:
                x_train_sampled, y_train_sampled = self.sampler[sampler_idx].sample(x_train, y_train)
            else:
                x_train_sampled, y_train_sampled = self.sampler[sampler_idx].sample(x_train_sampled, y_train_sampled)

        return x_train_sampled, y_train_sampled


import pandas as pd
from pkg_resources import resource_filename


def get_plz_mapping() -> pd.DataFrame:
    """
    get dataframe with ort-postleitzahl-landkreis-bundesland mapping

    @soruce: https://www.suche-postleitzahl.org/downloads, 18/07/2023
    """
    filepath = resource_filename(__name__, 'zuordnung_plz_ort.csv')
    df = pd.read_csv(filepath, dtype={'plz': str})
    df =  df[["ort", "plz", "landkreis", "bundesland"]]
    return df

def get_coord_main_cities() -> dict:
    """
    get coordinates of top cities from germany
    """
    top_cities = {
        'Berlin': (13.404954, 52.520008), 
        'Cologne': (6.953101, 50.935173),
        'Düsseldorf': (6.782048, 51.227144),
        'Frankfurt am Main': (8.682127, 50.110924),
        'Hamburg': (9.993682, 53.551086),
        'Leipzig': (12.387772, 51.343479),
        'Munich': (11.576124, 48.137154),
        'Dortmund': (7.468554, 51.513400),
        'Stuttgart': (9.181332, 48.777128),
        'Nuremberg': (11.077438, 49.449820),
        'Hannover': (9.73322, 52.37052)
    }
    return top_cities



import os

import pandas as pd
import pytest
from sklearn.datasets import make_classification
from sklearn.model_selection import train_test_split

from sam_ml.data.preprocessing import Sampler, Scaler, Selector

os.environ["SAM_ML_SOUND_ON"] = "False"
os.environ["SAM_ML_LOG_LEVEL"] = "debug"
from sam_ml.models import CTest

X, Y = make_classification(n_samples = 50,
                            n_features = 5,
                            n_informative = 5,
                            n_redundant = 0,
                            n_classes = 3,
                            weights = [.2, .3, .8])
X = pd.DataFrame(X, columns=["col1", "col2", "col3", "col4", "col5"])
Y = pd.Series(Y)
x_train, x_test, y_train, y_test = train_test_split(X,Y, train_size=0.80, random_state=42)


def test_eval_models_selectors():
    selectors = Selector.params()["algorithm"]
    ctest = CTest(models="all", selector=selectors)
    ctest.eval_models(x_train, y_train, x_test, y_test)

def test_eval_models_scalers():
    scalers = Scaler.params()["scaler"]
    ctest = CTest(models="all", scaler=scalers)
    ctest.eval_models(x_train, y_train, x_test, y_test)

def test_eval_models_cv():
    ctest = CTest(models="all")
    ctest.eval_models_cv(X, Y)

def test_eval_models_cv_small_data():
    ctest = CTest(models="all")
    ctest.eval_models_cv(X, Y, small_data_eval=True)

def test_find_best_model_randomCV():
    ctest = CTest(models="all")
    ctest.find_best_model_randomCV(x_train, y_train, x_test, y_test)

def test_find_best_model_randomCV_small_data():
    ctest = CTest(models="all")
    ctest.find_best_model_randomCV(x_train, y_train, x_test, y_test, small_data_eval=True)

def test_find_best_model_mass_search_error():
    with pytest.raises(RuntimeError):
        ctest = CTest(models="all")
        ctest.find_best_model_mass_search(x_train, y_train, x_test, y_test)

##### larger data set

X_200, Y_200 = make_classification(n_samples = 200,
                            n_features = 5,
                            n_informative = 5,
                            n_redundant = 0,
                            n_classes = 3,
                            weights = [.2, .3, .8])
X_200 = pd.DataFrame(X_200, columns=["col1", "col2", "col3", "col4", "col5"])
Y_200 = pd.Series(Y_200)
x_train_200, x_test_200, y_train_200, y_test_200 = train_test_split(X_200,Y_200, train_size=0.80, random_state=42)

X_5500, Y_5500 = make_classification(n_samples = 5500,
                            n_features = 5,
                            n_informative = 5,
                            n_redundant = 0,
                            n_classes = 3,
                            weights = [.2, .3, .8])
X_5500 = pd.DataFrame(X_5500, columns=["col1", "col2", "col3", "col4", "col5"])
Y_5500 = pd.Series(Y_5500)
x_train_5500, x_test_5500, y_train_5500, y_test_5500 = train_test_split(X_5500,Y_5500, train_size=0.80, random_state=42)


def test_eval_models_samplers():
    samplers = Sampler.params()["algorithm"]
    ctest = CTest(models="all", sampler=samplers)
    ctest.eval_models(x_train_200, y_train_200, x_test_200, y_test_200)

def test_find_best_model_mass_search():
    ctest = CTest(models="all")
    ctest.find_best_model_mass_search(x_train_5500, y_train_5500, x_test_5500, y_test_5500, n_trails=3)



from .global_variables import get_n_jobs, get_sound_on
from .logging import setup_logger

__all__ = {
    "logger setup function": "setup_logger",
    "get n_jobs parameter": "get_n_jobs",
    "get if sound on": "get_sound_on",
}
